/* $Workfile: T8pBurstEx.c $       $Revision: 39 $      $Modtime: 11/09/05 1:23p $  */
/*******************************************************************************

  Odin TeleSystems Inc.

  OTX Telecom Adapter Driver

  @doc PublicDemo

  @module

    T8pBurstEx.c

    A simple example console application demonstrating the use of the
    DMA Direct Sending and Receiving with the OTX Driver using the
    board's Burst device to access the data using DMA.

      Thor-2-PCI-Plus
      Thor-4-PCI-Plus
      Thor-8-PCI-Plus
      Sleipnir-1-PCI-Plus
      Hugin-2-PC104

    The Demo application provides a simple console interface which utilizes
    simple keystrokes as commands.

    It configures 2(4,8) Lines respectively for Thor-2(4,8)-PCI-Plus which can 
    be looped with an external loop cable.  Sleipnir-1-PCI-Plus can be looped 
    back to itself.

    The direct sends and receives operate on a highway basis and are
    hosted by the Burst device.

    Press 'h' or '?' for help. 'q' to quit.

*******************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <assert.h>

#include "OsaStd.h"
#include "OsaCon.h"
#include "OsaThrd.h"
#include "OtxErr.h"
#include "OtxDef.h"
#include "OtxType.h"
#include "OtxDrv.h"
#include "OtxHdlc.h"
#include "OtxRaw.h"
#include "OtxT1E1.h"
#include "OtxData.h"
#include "OtxTone.h"
#include "OtxDsp.h"
#include "OtxBrd.h"

#include "OtxBurst.h"
#include "OtxBurstD.h"

#define DEMO_DISPLAY_BYTES 1   // Comments this macro out to display only the received F-bits (T1 mode only) in verbose mode

#define DEMO_DEFAULT_LOG_FILE   "T8pBurstEx.log"
#define DEMO_DEFAULT_BOARD_NO   0

#define DEMO_BURST_BUF_SIZE     0x8000    // 32k


typedef struct DemoEnvT {
  char            *pszInputStr;
  OTX_BOOL        bExitOnError;
  char            cwd[OTX_MAX_PATH];
  char            logFileName[OTX_MAX_PATH];
  OtxPhysicalDeviceTypeE nBoardType;
  int             coLi;
  int             coHighways;
  int             nCurrentLi;
  OtxT1E1LiModeE  eT1E1Mode;    // T1 or E1 mode
  int             boardNo;
  OTX_BOOL        bDisableBurstTx;
  OTX_BOOL        bDisableBurstRx;
} DemoEnvT;


/*******************************************************************************
 Function Prototypes
*******************************************************************************/
static void DemoCreateEventQueue(void);
static void DemoCreatePhysicalDevices(void);
static void DemoConfigureE1T1Devices(void);
static void DemoEnablePhysicalDevices(void);
static void DemoConnectDevices(void);
static void DemoInitT1E1ConfigData(OUT OtxE1ConfigOptionsS *pE1CfgData, OUT OtxT1ConfigOptionsS *pT1CfgData);
static void DemoPrintEvent(IN OtxEventDataS *pEventData);
static void DemoCheckRc(IN OTX_HANDLE hDevice, IN OTX_RESULT otxRc);
static void DemoExitQuestion(void);
static void DemoLogf(char *formatStr, ...);
static void DemoHelpPanel(void);
static void DemoParseCmdLineArgs(int argc, char *argv[], DemoEnvT *envP);
static void DemoBurstSender(OTX_UINT32 nSourceId);
static void DemoBurstReceiver(OTX_UINT32 nSourceId);


/*******************************************************************************
 Global Data
*******************************************************************************/
OTX_EVENT   g_hNotificationEvent = OTX_INVALID_EVENT_VALUE;
OTX_HANDLE  g_hEventQueue  = 0;
FILE        *g_pDemoLogFile = NULL;
DemoEnvT    g_demoEnv;
OTX_BOOL    g_bStatsThreadActive = OTX_FALSE;


struct {
  OTX_HANDLE hThorController;
  OTX_HANDLE hLi[8];
  OTX_HANDLE hQuadLi[2];
  OTX_HANDLE hBurst;
} g_hPhysDevs;

struct {
  OTX_HANDLE hLiIn[8];
  OTX_HANDLE hLiOut[8];
} g_hPipes;

// direct Burst DMA send
struct sBurstSender {
  OTX_UINT32  nSourceId;
  OTX_BOOL    bVerbose;
  OTX_BOOL    bActive;
  OTX_UINT8   sqTxBuf[DEMO_BURST_BUF_SIZE];
  FILE        *pTxFile;
  OTX_INT32   nTsToSend;  // Default is -1 (All TS)
  OTX_UINT32  cbTxBuf;
  OTX_BOOL    bFileLoop;
};

// direct Burst DMA receive
struct sBurstReceiver {
  OTX_UINT32  nSourceId;
  OTX_BOOL    bVerbose;
  OTX_BOOL    bActive;
  FILE        *pRxFile;
  OTX_INT32   nTsToRecord;  // Default is -1 (All TS)
};

struct {
  struct sBurstSender mBurstSender[2];        // one for each possible highway hosted by the Burst device
  struct sBurstReceiver mBurstReceiver[2];    // one for each possible highway hosted by the Burst device
} g_hLogicDevs;

struct {
  OtxE1ConfigOptionsS e1CfgData;
  OtxT1ConfigOptionsS t1CfgData;
} g_T1E1CfgData;




/*******************************************************************************
*******************************************************************************/


void DemoCreateEventQueue(void)
/*******************************************************************************

  @func

  Create the event queue to be used by the driver to inform the application
  about events, such as line status changes, or detected DMTF digits.

*******************************************************************************/
{
  OTX_RESULT result;

  // Create a notification Event
  result = OtxDrvCreateEvent(&g_hNotificationEvent);
  if (OTX_RESULT_CODE(result) == OTX_E_NOTIMPL) {
    g_hNotificationEvent = NULL;
    result = OTX_S_OK;
  }

  DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);

  // Create Event Queue for driver notifications
  result = OtxDrvCreateEventQueue(128, g_hNotificationEvent, &g_hEventQueue); DemoCheckRc(g_hEventQueue, result);

  // Enable Event Queue
  result = OtxDrvEnable(g_hEventQueue, OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hEventQueue, result);

}



void DemoCreatePhysicalDevices(void)
/*******************************************************************************

  @func

  Create and Connect the Physical Devices on the Thor-8 board

  @comm

    Calls function <f OtxDrvCreatePhysicalDevice> returning a handle to a
    Device Object representing a physical device, which can be the
    entire board or an integrated circuit (chip) on the board. The number of
    physical devices is set per board and cannot change dynamically.

    The physical devices can be classified in 2 categories:

  - BOARD Devices: A board device represents one OTX Network Interface card or
    a Daughter Board Module. The board device is the place holder and controller
    of other physical and sometime logical devices representing
    the devices on the board: such as Line Interface and Filter  Devices.

    The Burst device hosts the DMA functionality.  If we open the Burst
    device, when we enable that device, the process of transferring data from
    the board to the driver will take place via DMA.  The Burst device provides
    the necessary buffering to process the data from the board until it can be
    processed and filtered to the logical devices that are hosted by Burst.
    (see DemoCreateLogicalDevices for example of connection).

  - CHIP Devices: A chip device represents a physical integrated circuit
    on the board performing one or several functions. A logical device can
    be created to represent each of the functions implemented by a chip device.
    For example, a Digital Signal Processor is a physical "chip" device.
    However, the DSP software may implement several functions and be
    represented by a multiple of logical devices.



*******************************************************************************/
{
  OTX_RESULT result;
  OTX_INT32 i;
  OTX_INT32 j;

  // Connect to the Physical Board Controller
  switch (g_demoEnv.nBoardType) {
    case OTX_DEVICE_SLEIPNIR_1_PCI:
      DemoLogf("Opening Sleipnir board\n");
      break;
    case OTX_DEVICE_HUGIN_2_PC104:
      DemoLogf("Opening Hugin-2-PC104 board\n");
      break;
    case OTX_DEVICE_THOR_2_PCI_PLUS:
      DemoLogf("Opening Thor-2-PCI Plus board\n");
      break;
    case OTX_DEVICE_THOR_4_PCI_PLUS:
      DemoLogf("Opening Thor-4-PCI Plus board\n");
      break;
    case OTX_DEVICE_THOR_8_PCI_PLUS:
      DemoLogf("Opening Thor-8-PCI Plus board\n");
      break;
    case OTX_DEVICE_THOR_1_PCI_PLUS:
      DemoLogf("Opening Thor-1-PCI Plus board\n");
      break;
    default:
      DemoLogf("Board Type (%d) not supported by this demo program\n", g_demoEnv.nBoardType);
      OsaConsoleExit();
      exit(1);
      break;
  }

  result = OtxDrvOpenPhysicalDevice(0, g_demoEnv.nBoardType, g_demoEnv.boardNo, g_demoEnv.boardNo, g_hEventQueue, &g_hPhysDevs.hThorController); DemoCheckRc(g_hPhysDevs.hThorController, result);

  if (g_demoEnv.nBoardType == OTX_DEVICE_THOR_2_PCI_PLUS) {
    for (i = 0; i < 2; i++) {
      // Get handles to the 2 individual T1/E1 line interface FALC 
      result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hThorController, OTX_DEVICE_LI_T1E1, i, i, g_hEventQueue, &g_hPhysDevs.hLi[i]); DemoCheckRc(g_hPhysDevs.hThorController, result);
    }

  } else if (g_demoEnv.nBoardType == OTX_DEVICE_THOR_4_PCI_PLUS) {
    for (i = 0; i < 1; i++) {
      // Get handles to the one quad T1/E1 line interface devices
      result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hThorController, OTX_DEVICE_QUAD_T1E1, i, i, g_hEventQueue, &g_hPhysDevs.hQuadLi[i]); DemoCheckRc(g_hPhysDevs.hThorController, result);

      // Get handles to the 4 individual T1/E1 line interface members within the Quad device
      for (j = 0; j < 4; j++) {
        result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hQuadLi[i], OTX_DEVICE_LI_T1E1, j, j + i*4, g_hEventQueue, &g_hPhysDevs.hLi[j + i*4]); DemoCheckRc(g_hPhysDevs.hQuadLi[i], result);
      }
    }

  } else if (g_demoEnv.nBoardType == OTX_DEVICE_THOR_8_PCI_PLUS) {
    for (i = 0; i < 2; i++) {
      // Get handles to the two quad T1/E1 line interface devices
      result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hThorController, OTX_DEVICE_QUAD_T1E1, i, i, g_hEventQueue, &g_hPhysDevs.hQuadLi[i]); DemoCheckRc(g_hPhysDevs.hThorController, result);

      // Get handles to the 4 individual T1/E1 line interface members within the Quad device
      for (j = 0; j < 4; j++) {
        result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hQuadLi[i], OTX_DEVICE_LI_T1E1, j, j + i*4, g_hEventQueue, &g_hPhysDevs.hLi[j + i*4]); DemoCheckRc(g_hPhysDevs.hQuadLi[i], result);
      }
    }

  } else if (g_demoEnv.nBoardType == OTX_DEVICE_SLEIPNIR_1_PCI || g_demoEnv.nBoardType == OTX_DEVICE_HUGIN_2_PC104) {

    // Open the Line interfaces
    for (i = 0; i < g_demoEnv.coLi; i++) {
      result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hThorController, OTX_DEVICE_LI_T1E1, i, i, g_hEventQueue, &g_hPhysDevs.hLi[i]); DemoCheckRc(g_hPhysDevs.hThorController, result);
    }

    // we'll source the Falc from the Burst device, and source the burst device from the FALC
    //  Falc <-- Burst Send (P2S)
    result = OtxDrvSetAttributeValueUint32(g_hPhysDevs.hThorController, OTX_ATTR_BRD_FALC_SOURCED, OTX_BRD_FALC_SOURCED_P2S); DemoCheckRc(g_hPhysDevs.hThorController, result);
    //  Burst Recv (S2P) <-- Falc
    result = OtxDrvSetAttributeValueUint32(g_hPhysDevs.hThorController, OTX_ATTR_BRD_S2P_SOURCED, OTX_BRD_S2P_SOURCED_FALC); DemoCheckRc(g_hPhysDevs.hThorController, result);
  
  } else if (g_demoEnv.nBoardType == OTX_DEVICE_THOR_1_PCI_PLUS) {
    for (i = 0; i < 1; i++) {
      // Get handles to individual T1/E1 line interface FALC 
      result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hThorController, OTX_DEVICE_LI_T1E1, i, i, g_hEventQueue, &g_hPhysDevs.hLi[i]); DemoCheckRc(g_hPhysDevs.hThorController, result);
    }

  }

  // Get a handle to the physical Burst device.
  result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hThorController, OTX_DEVICE_BURST, 0, 0, g_hEventQueue, &g_hPhysDevs.hBurst); DemoCheckRc(g_hPhysDevs.hThorController, result);

  // set device buffer event trigger levels
  result = OtxDrvSetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_RECEIVER_FIFO_TRIGGER,  DEMO_BURST_BUF_SIZE); DemoCheckRc(g_hPhysDevs.hBurst, result);
  result = OtxDrvSetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_SENDER_FIFO_TRIGGER,    DEMO_BURST_BUF_SIZE); DemoCheckRc(g_hPhysDevs.hBurst, result);

  // Optionally (mainly for testing, or for improving system performance for single direction burst applications,
  // disable one direction of the burst device
  if (g_demoEnv.bDisableBurstRx) {
    DemoLogf("Burst RX interrupts will be turned off when the Burst device is enabled\n");
    result = OtxDrvSetAttributeValueBool(g_hPhysDevs.hBurst, OTX_ATTR_BURST_RECV_DISABLED,  OTX_TRUE); DemoCheckRc(g_hPhysDevs.hBurst, result);
  }
  if (g_demoEnv.bDisableBurstTx) {
    DemoLogf("Burst TX interrupts will be turned off when the Burst device is enabled\n");
    result = OtxDrvSetAttributeValueBool(g_hPhysDevs.hBurst, OTX_ATTR_BURST_SEND_DISABLED,  OTX_TRUE); DemoCheckRc(g_hPhysDevs.hBurst, result);
  }
}



/*******************************************************************************

  @func
    Configure E1/T1 Devices

  @comm The clock source of the board(s) must set by a call
  to <f OtxBrdSetClocks> prior to calling this function.

*******************************************************************************/
static void DemoConfigureE1T1Devices(void)
{
  OTX_RESULT result;
  OTX_INT32  i;

  // Set default configuration data
  DemoInitT1E1ConfigData(&(g_T1E1CfgData.e1CfgData), &(g_T1E1CfgData.t1CfgData));

  if (g_demoEnv.eT1E1Mode == OTX_T1E1_MODE_T1){ // T1
    // Configure T1 Transceivers
    for (i = 0; i < g_demoEnv.coLi; i++) {
      result = OtxT1SetConfiguration(g_hPhysDevs.hLi[i], &(g_T1E1CfgData.t1CfgData));  DemoCheckRc(g_hPhysDevs.hLi[i], result);
    }
    // Set them in T1 mode
    for (i = 0; i < g_demoEnv.coLi; i++) {
      result =  OtxT1E1SetLiMode(g_hPhysDevs.hLi[i], OTX_T1E1_MODE_T1); DemoCheckRc(g_hPhysDevs.hLi[i], result);
    }
  }
  else{  // E1
    // Configure E1 Transceivers
    for (i = 0; i < g_demoEnv.coLi; i++) {
      result = OtxE1SetConfiguration(g_hPhysDevs.hLi[i], &(g_T1E1CfgData.e1CfgData));  DemoCheckRc(g_hPhysDevs.hLi[i], result);
    }
    // Set them in E1 mode
    for (i = 0; i < g_demoEnv.coLi; i++) {
      result =  OtxT1E1SetLiMode(g_hPhysDevs.hLi[i], OTX_T1E1_MODE_E1); DemoCheckRc(g_hPhysDevs.hLi[i], result);
    }
  }

}


void DemoEnablePhysicalDevices(void)
/*******************************************************************************

  @func
    Enable Configured Physical Devices

*******************************************************************************/
{
  OTX_RESULT result;
  OTX_INT32 i;

  // Enable the board (Physical Device)
  DemoLogf("Enable g_hPhysDevs.hThorController\n");
  result = OtxDrvEnable(g_hPhysDevs.hThorController,  OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hPhysDevs.hThorController, result);

  if (g_demoEnv.nBoardType == OTX_DEVICE_THOR_4_PCI_PLUS) {

    for (i = 0; i < 1; i++) {
      // Enable QuadFalcs
      DemoLogf("Enable g_hPhysDevs.hQuadLi[%d]\n", i);
      result = OtxDrvEnable(g_hPhysDevs.hQuadLi[i],  OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hPhysDevs.hQuadLi[i], result);
    }

  }

  if (g_demoEnv.nBoardType == OTX_DEVICE_THOR_8_PCI_PLUS) {

    for (i = 0; i < 2; i++) {
      // Enable QuadFalcs
      DemoLogf("Enable g_hPhysDevs.hQuadLi[%d]\n", i);
      result = OtxDrvEnable(g_hPhysDevs.hQuadLi[i],  OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hPhysDevs.hQuadLi[i], result);
    }

  }

  for (i = 0; i < g_demoEnv.coLi; i++) {
    // Enable Physical Devices, Li's
    DemoLogf("Enable g_hPhysDevs.hLi[%d]\n", i);
    result = OtxDrvEnable(g_hPhysDevs.hLi[i],  OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hPhysDevs.hLi[i], result);
  }

}


void DemoConnectDevices(
/*******************************************************************************

  @func
    Create the needed Pipe Devices and connect the physical and logical devices
    to pipes.

  @comm
    A Pipe Device is a logical entity representing a serial bit-stream (a subset of a
    Highway) with a data rate between 8 kbit/s and 2.048Mbit/s. A pipe is a characterized
    bu its capacity (in 8 kbit/s increments). The user can create a pipe with
    the driver function <f OtxDrvCreatePipe>. A pipe is connected to physical or
    logical devices and mapped to specific time-slots with the driver function
    <f OtxDrvConnectPipe>.

*******************************************************************************/
  void
)
{
  OTX_INT32 nLi;
  OTX_RESULT result;
  OTX_UINT32 iHighwayNo;
  OTX_UINT8 sqTimeSlotMask[32];
  OTX_UINT8 sqBurstTimeSlotMask[OTX_PIPE_TIMESLOTS_MAX];


  // Create the Pipes (one for each direction) between the Burst Device and the LIs
  for (nLi = 0; nLi < g_demoEnv.coLi; nLi++) {
    OTX_INT32 nTs;

    // we must define the highway to use [0-1]
    iHighwayNo = nLi / (g_demoEnv.coLi / g_demoEnv.coHighways);

    // Create pipes (one for each direction) Use all timeslots.
    memset(sqTimeSlotMask, 0, sizeof(sqTimeSlotMask) * sizeof(sqTimeSlotMask[0]));
    memset(sqBurstTimeSlotMask, 0, sizeof(sqBurstTimeSlotMask) * sizeof(sqBurstTimeSlotMask[0]));

    DemoLogf("Creating pipe for Burst Sender/Receiver\n");

    for (nTs = 0; nTs < 32; nTs++) {
      
      sqTimeSlotMask[nTs] = 0xFF;
      
      // for this demo, we'll lay out the burst time slot mask as follows:
      // OTX_DEVICE_SLEIPNIR_1_PCI:
      //    [ 0..31]  - LI0:TS0-TS31
      //
      // OTX_DEVICE_HUGIN_2_PC104:
      // OTX_DEVICE_THOR_2_PCI_PLUS:
      //    [ 0..31]  - LI0:TS0-TS31 (hwy0) and LI1:TS0-TS31 (hwy1) 
      // 
      // OTX_DEVICE_THOR_8_PCI_PLUS:
      //    [ 0..31]  - LI0:TS0-TS31 (hwy0)  and LI4:TS0-TS31 (hwy1) 
      //    [32..63]  - LI1:TS0-TS31 (hwy0)  and LI5:TS0-TS31 (hwy1) 
      //    [64..95]  - LI2:TS0-TS31 (hwy0)  and LI6:TS0-TS31 (hwy1) 
      //    [96..127] - LI3:TS0-TS31 (hwy0)  and LI7:TS0-TS31 (hwy1) 
      //
      // OTX_DEVICE_GIMLE_16_PCI_PLUS:
      //    [ 0..31]  - LI0:TS0-TS31 (hwy0)  and LI4:TS0-TS31 (hwy1)  and LI8:TS0-TS31  (hwy2)  and LI8:TS0-TS31  (hwy3)
      //    [32..63]  - LI1:TS0-TS31 (hwy0)  and LI5:TS0-TS31 (hwy1)  and LI9:TS0-TS31  (hwy2)  and LI9:TS0-TS31  (hwy3)
      //    [64..95]  - LI2:TS0-TS31 (hwy0)  and LI6:TS0-TS31 (hwy1)  and LI10:TS0-TS31 (hwy2)  and LI10:TS0-TS31 (hwy3)
      //    [96..127] - LI3:TS0-TS31 (hwy0)  and LI7:TS0-TS31 (hwy1)  and LI11:TS0-TS31 (hwy2)  and LI11:TS0-TS31 (hwy3)
      //
      // For OTX_DEVICE_THOR_8_PCI_PLUS and OTX_DEVICE_GIMLE_16_PCI_PLUS 
      // upon connecting the pipe (see below), we have to pass the highway
      // number (0,1 for Thor-8 and 0,1,2,3 for Gimle-16). There is one highway 
      // per each QuadFalc (128TS per highway). We also pass the size of the highway,
      // which is 128 (TS) for these two boards.
      //
      // For OTX_DEVICE_SLEIPNIR_1_PCI, OTX_DEVICE_HUGIN_2_PC104, and 
      // OTX_DEVICE_THOR_2_PCI_PLUS, the size of the highway is 32.
      //
      // this is just a simple algorithm to accomplish the above
      sqBurstTimeSlotMask[nTs + ((nLi % (g_demoEnv.coLi / g_demoEnv.coHighways)) * 32)] = 0xFF;
      
    }


    // --------------------------------------------------------------------
    // Create an incoming and outgoing Pipe representing the external T1/E1
    // --------------------------------------------------------------------
    result = OtxDrvCreatePipe(64*32, &g_hPipes.hLiIn[nLi]);
    DemoCheckRc(g_hPipes.hLiIn[nLi], result);
    result = OtxDrvCreatePipe(64*32, &g_hPipes.hLiOut[nLi]);
    DemoCheckRc(g_hPipes.hLiOut[nLi], result);

    // -------------------------
    // Connect pipe: Burst -> Li
    // -------------------------
    result = OtxDrvConnectPipe(g_hPipes.hLiIn[nLi], OTX_PIPE_INPUT,  g_hPhysDevs.hBurst, iHighwayNo, sqBurstTimeSlotMask, ((g_demoEnv.coLi / g_demoEnv.coHighways) * 32));
    DemoCheckRc(g_hPipes.hLiIn[nLi], result);

    result = OtxDrvConnectPipe(g_hPipes.hLiIn[nLi], OTX_PIPE_OUTPUT, g_hPhysDevs.hLi[nLi],  -1, sqTimeSlotMask, 32);
    DemoCheckRc(g_hPipes.hLiIn[nLi], result);

    // -------------------------
    // Connect pipe: Li -> Burst
    // -------------------------
    result = OtxDrvConnectPipe(g_hPipes.hLiOut[nLi], OTX_PIPE_INPUT,  g_hPhysDevs.hLi[nLi],  -1, sqTimeSlotMask, 32);
    DemoCheckRc(g_hPipes.hLiOut[nLi], result);

    result = OtxDrvConnectPipe(g_hPipes.hLiOut[nLi], OTX_PIPE_OUTPUT, g_hPhysDevs.hBurst, iHighwayNo, sqBurstTimeSlotMask, ((g_demoEnv.coLi / g_demoEnv.coHighways) * 32));
    DemoCheckRc(g_hPipes.hLiOut[nLi], result);

  }

}


void DemoInitT1E1ConfigData(
/*******************************************************************************

  @func
    Initialize the T1/E1 Configuration Data Structures passed to the T1/E1
    Line Interface physical devices

*******************************************************************************/
  OUT OtxE1ConfigOptionsS *pE1CfgData,
  OUT OtxT1ConfigOptionsS *pT1CfgData
)
{
  // Initialize E1 Configuration Data
  pE1CfgData->eTxLineCode                = OTX_E1_LC_HDB3;
  pE1CfgData->eRxLineCode                = OTX_E1_LC_HDB3;
  pE1CfgData->eTxFrameFormat             = OTX_E1_FF_CRC4_MULTIFRAME;
  pE1CfgData->eRxFrameFormat             = OTX_E1_FF_CRC4_MULTIFRAME;
  pE1CfgData->eAisDetection              = OTX_E1_AIS_ETS300233;
  pE1CfgData->bfSiBits                   = 0x3;
  pE1CfgData->bfSaBits                   = 0x1F;
  pE1CfgData->bExtendedHDB3errorDetect   = OTX_FALSE;
  pE1CfgData->bAutoRegainMultiframe      = OTX_FALSE;
  pE1CfgData->bAutoResync                = OTX_TRUE;
  pE1CfgData->bAutoCrcBitInversion       = OTX_FALSE;
  pE1CfgData->bAutoRemoteAlarm           = OTX_FALSE;
  pE1CfgData->nLosSensitivity            = 10;  // 10 = LOS detection after 176 consecutive zeros which fulfills G.775
  pE1CfgData->nLosRecovery               = 21;  // 21 = LOS recovery after 22 ones in the LosRecovery interval which fulfills G.775
  pE1CfgData->nLineLength                = 10;
  pE1CfgData->bHighTxPower               = OTX_FALSE;
  pE1CfgData->bReceiveEqualizer          = OTX_FALSE;

  // Initialize T1 Configuration Data
  pT1CfgData->eSignalingMode             = OTX_T1_SM_CCS;
  pT1CfgData->eTxLineCode                = OTX_T1_LC_B8ZS;
  pT1CfgData->eRxLineCode                = OTX_T1_LC_B8ZS;
  pT1CfgData->eFrameFormat               = OTX_T1_FF_ESF;
  pT1CfgData->bCRC6Enable                = OTX_TRUE;
  pT1CfgData->eTxRemoteAlarmFormat       = OTX_T1_RA_YELLOW_A;
  pT1CfgData->eRxRemoteAlarmFormat       = OTX_T1_RA_YELLOW_A;
  pT1CfgData->bAutoResync                = OTX_TRUE;
  pT1CfgData->bAutoCrcBitInversion       = OTX_FALSE;
  pT1CfgData->eLfaSensitivity            = OTX_T1_LFA_2_OUT_OF_4;
  pT1CfgData->bAutoRemoteAlarm           = OTX_FALSE;
  pT1CfgData->nLosSensitivity            = 10;  // 10 = LOS detection after 176 consecutive zeros which fulfills G.775, Telcordia (Bellcore) and AT&T
  pT1CfgData->nLosRecovery               = 21;  // 21 = LOS recovery after 22 ones in the LosRecovery interval which fulfills G.775, Telcordia (Bellcore) and AT&T
  pT1CfgData->nLineLength                = 10;
  pT1CfgData->bHighTxPower               = OTX_FALSE;
  pT1CfgData->bReceiveEqualizer          = OTX_TRUE;
  pT1CfgData->bfClearChannels            = 0x000000; // 0x000000 = No channels are marked as clear channels
}


/*******************************************************************************

  @func

  Displays a message on the screen 

  @comm

*******************************************************************************/
static void DemoPrintf(char *formatStr, ...)
{
  va_list argList;
  va_start(argList, formatStr);

  OsaConsolePrintF(formatStr, argList);

  va_end(argList);

}

static void DemoCheckRc(IN OTX_HANDLE hDevice, IN OTX_RESULT result)
/*******************************************************************************
  @func
    Check the result from API call. If the result is a failure code then prompt
    the user if he wants to exit.
*******************************************************************************/
{
  if ( OTX_FAILURE(result) ) {

    if (hDevice == OTX_INVALID_HANDLE_VALUE) {
      DemoLogf("ERROR(%#lX) \n", result);
    } else {
      OTX_RESULT otxRc;
      char msgStr[256];
      OtxAttribTypeE  eAttribType;    // Type of the attribute, if attribute exists
      OTX_UINT32      cbAttribValue;  // Size in bytes of attribute value (Get value with <f OtxDrvGetAttributeValue>)
      OTX_BOOL        bReadOnly;      // Is attribute writable or read-only
      OTX_UINT32      eDeviceType = 999;    // Device type for hDevice
      OTX_UINT32      nDeviceNumber = 999;  // Device number for hDevice

      // Get the Device Type for hHandle
      otxRc = OtxDrvGetAttributeType(hDevice, OTX_ATTR_DEVICE_TYPE, &eAttribType, &cbAttribValue, &bReadOnly);
      if (OTX_SUCCESS(otxRc)) {
        assert(eAttribType == OTX_ATTRIB_UINT32);
        assert(cbAttribValue == sizeof(OTX_UINT32));
        OtxDrvGetAttributeValueUint32(hDevice, OTX_ATTR_DEVICE_TYPE, &eDeviceType);
      }

      // Get the Device Number for hHandle
      otxRc = OtxDrvGetAttributeType(hDevice, OTX_ATTR_DEVICE_NO, &eAttribType, &cbAttribValue, &bReadOnly);
      if (OTX_SUCCESS(otxRc)) {
        assert(eAttribType == OTX_ATTRIB_UINT32);
        assert(cbAttribValue == sizeof(OTX_UINT32));
        OtxDrvGetAttributeValueUint32(hDevice, OTX_ATTR_DEVICE_NO, &nDeviceNumber);
      }

      // Convert the result code to a more descriptive string
      otxRc = OtxDrvResultCode2String(hDevice, result, 256, &msgStr[0]);
      if (OTX_FAILURE(otxRc)) {
        strcpy(msgStr, "Cannot translate this result code.");
      }

      DemoLogf("ERROR(%#lX [%d.%d]) Dev(%X,%d,%d): %s \n", OTX_RESULT_CODE(result), OTX_RESULT_SOURCE_ID(result), OTX_RESULT_LINE_NUM(result), hDevice, (int) eDeviceType, (int)nDeviceNumber, msgStr);
#if 0
      {
        // API call to get the result code data
        OTX_UINT32 Result, Severity, SourceId, LineNum;

        OtxResultCode(result, &Result, &Severity, &SourceId, &LineNum);
        DemoLogf("ERROR(%#lX [%d.%d]) Dev(%X,%d,%d): %s \n", Result, SourceId, LineNum, hDevice, (int) eDeviceType, (int)nDeviceNumber, msgStr);
      }
#endif

    }

    DemoExitQuestion();
  }
}


/*******************************************************************************
  @func
    Prompts whether to exit the program. Usually called after an error
    has been reported.

*******************************************************************************/
void DemoExitQuestion()
{
  short c;

  if (g_demoEnv.bExitOnError) {
    c = 'y';
  } else {
    DemoLogf("Do you want to exit (Y/N) ");
    c = OsaConsoleGetCh();
  }
  if (c == 'y' || c == 'Y') {
    // Clean up
    OtxDrvReleaseLib();
    OsaConsoleExit();
    exit(1);
  }
}


/*******************************************************************************
  @func
    Prints a timestamp in a MM-DD hh:mm:ss:mmm format, where

    MM  = month
    DD  = day (of month)
    hh  = hour of day
    mm  = minutes of day
    ss  = seconds of day
    mmm = milliseconds

*******************************************************************************/
void DemoPrintTimeStamp(
  OTX_DATETIME nTimeStamp
)
{
  OTX_DATETIME nTimeStampMs;
  OTX_DATETIME nTimeStampSec;
  struct Osa_tm *ptCurrentDayAndTime;
  Osa_time_t nCurrentTimeSecs;

  // Subtract number of 100ns ticks between Jan 1, 1970 and Jan 1, 1600
  #if defined(__linux__)
    nTimeStamp.nUint64 = nTimeStamp.nUint64 - 116444736000000000LL;  
  #elif defined(WIN32)
    nTimeStamp.nUint64 = nTimeStamp.nUint64 - 116444736000000000;
  #endif

  // Convert to seconds since Jan 1, 1970 (so we can use this time with OsaStdLocalTime())
  nTimeStampSec.nUint64 = nTimeStamp.nUint64 / 10000000;

  // Calculate number of milliseconds (since number of seconds is not a very accurate way to report timestamp)
  nTimeStampMs.nUint64 = nTimeStamp.nUint64 / 10000 - (nTimeStampSec.nUint64 * 1000);

  nCurrentTimeSecs = nTimeStampSec.sUint32.nLow;
  ptCurrentDayAndTime = OsaStdLocalTime(&nCurrentTimeSecs);

  if (ptCurrentDayAndTime) {

    DemoLogf("%.2d-%.2d %.2d:%.2d:%.2d.%.3d ",
      ptCurrentDayAndTime->tm_mon + 1,
      ptCurrentDayAndTime->tm_mday,
      ptCurrentDayAndTime->tm_hour,
      ptCurrentDayAndTime->tm_min,
      ptCurrentDayAndTime->tm_sec,
      nTimeStampMs.sUint32.nLow
      );
  } else {
    DemoLogf("Error: Invalid timestamp (before Jan 1, 1970)\n");
  }

}


/*******************************************************************************
  @func
    Idle Function that check for Events from the OTX Driver
*******************************************************************************/
void *DemoIdleFunction(
  void *lpParameter   // thread data
)
{
  OtxEventDataS eventData;
  OTX_RESULT result;

  while (1) {
    OtxDrvWaitForSingleEvent(g_hNotificationEvent, OTX_TIME_INFINITY);

    do {
      result = OtxDrvGetEventData(g_hEventQueue, &eventData); DemoCheckRc(g_hEventQueue, result);
      if (OTX_RESULT_CODE(result) == OTX_S_OK) {
        // We have a driver event

        switch (eventData.m_eDeviceType) {
          case OTX_DEVICE_BURST:
            switch (eventData.m_nCode) {
            case OTX_BURST_RECEIVER_EC_DATA_AVAILABLE:
              DemoBurstReceiver(eventData.m_nRequestId);
              continue;
            case OTX_BURST_SENDER_EC_BUF_DEPLETED:
              DemoBurstSender(eventData.m_nRequestId);
              continue;
            default:
              break;
            }
            break;
          default:
            break;
        }

        // Printf the event data.
        // In a real application branch based on the data
        DemoPrintEvent(&eventData);
      }
    } while (OTX_RESULT_CODE(result) == OTX_S_OK);

  }

  return 0;
}


/*******************************************************************************
  @func
    Print DMA statistics.

    These stats are obtained from the driver, and are being used during
    initial release as a debugging feature.  These stats may be eliminated
    in the future.
*******************************************************************************/
void *DemoStatsThread(
  void *lpParameter   // thread data
)
{
  OTX_RESULT        result;
  OTX_DATETIME      nTimeStamp;
  OTX_UINT32        stat_Starts;            // Number of actual dma transfers initiated
  OTX_UINT32        stat_Completes;         // Number of actual dma transfers completed
  OTX_UINT32        stat_Partial;           // Number of partial transfer due to framemask
  OTX_UINT32        stat_Full;              // Number of count of max dma size transfers
  OTX_UINT32        stat_Aborts;            // Number of actual dma transfers aborts
  OTX_UINT32        stat_Amount;            // Number of amount of data transfered
  OTX_UINT32        stat_Errors;            // Number of count of processing errors
  OTX_UINT32        stat_OUruns;            // Number of amount of over/under runs
  OTX_UINT32        stat_OUrunsUser;        // Number of amount of over/under runs caused by user mode
  OTX_UINT32        stat_DmaBusy;           // Number of count of DMA busy occurances
  OTX_UINT32        stat_Notify;            // Number of logical device notifies
  OTX_UINT32        stat_NotifyPartial;     // Number of logical device notifies - partial
  OTX_UINT32        stat_FifoSize;          // size of the Burst FIFO
  OTX_UINT32        stat_FifoTrigger;       // the current FIFO Trigger value
  OTX_UINT32        stat_FifoCount;         // the current amount in the FIFO
  OTX_UINT32        stat_FifoPeak;          // the peak FIFO usage
  OTX_UINT32        stat_FifoTriggerCount;  // the current FIFO trigger fire value

#define STATS_SHOW_FRAME_REF
#ifdef STATS_SHOW_FRAME_REF
  OTX_UINT32                ui32;
  OtxBurstFrameRefAndTime   FrameRef;             // frame reference data
  static OTX_UINT64         nLastFrame = 0;
#endif  // #ifdef STATS_SHOW_FRAME_REF
  

  while (g_bStatsThreadActive) {
    OtxDrvGetCurrentTime(&nTimeStamp);
    
    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_STAT_RECV_STARTS,             &stat_Starts            ); DemoCheckRc(g_hPhysDevs.hBurst, result);
    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_STAT_RECV_COMPLETES,          &stat_Completes         ); DemoCheckRc(g_hPhysDevs.hBurst, result);
    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_STAT_RECV_PARTIAL,            &stat_Partial           ); DemoCheckRc(g_hPhysDevs.hBurst, result);
    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_STAT_RECV_FULL,               &stat_Full              ); DemoCheckRc(g_hPhysDevs.hBurst, result);
    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_STAT_RECV_ABORTS,             &stat_Aborts            ); DemoCheckRc(g_hPhysDevs.hBurst, result);
    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_STAT_RECV_AMOUNT,             &stat_Amount            ); DemoCheckRc(g_hPhysDevs.hBurst, result);
    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_STAT_RECV_ERRORS,             &stat_Errors            ); DemoCheckRc(g_hPhysDevs.hBurst, result);
    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_STAT_RECV_OURUNS,             &stat_OUruns            ); DemoCheckRc(g_hPhysDevs.hBurst, result);
    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_STAT_RECV_OURUNS_USER,        &stat_OUrunsUser        ); DemoCheckRc(g_hPhysDevs.hBurst, result);
    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_STAT_RECV_DMABUSY,            &stat_DmaBusy           ); DemoCheckRc(g_hPhysDevs.hBurst, result);
    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_STAT_RECV_NOTIFY,             &stat_Notify            ); DemoCheckRc(g_hPhysDevs.hBurst, result);
    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_STAT_RECV_NOTIFY_PARTIAL,     &stat_NotifyPartial     ); DemoCheckRc(g_hPhysDevs.hBurst, result);
    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_RECEIVER_FIFO_SIZE,           &stat_FifoSize          ); DemoCheckRc(g_hPhysDevs.hBurst, result);
    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_RECEIVER_FIFO_TRIGGER,        &stat_FifoTrigger       ); DemoCheckRc(g_hPhysDevs.hBurst, result);
    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_RECEIVER_FIFO_COUNT,          &stat_FifoCount         ); DemoCheckRc(g_hPhysDevs.hBurst, result);
    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_RECEIVER_FIFO_PEAK,           &stat_FifoPeak          ); DemoCheckRc(g_hPhysDevs.hBurst, result);
    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_RECEIVER_FIFO_TRIGGER_COUNT,  &stat_FifoTriggerCount  ); DemoCheckRc(g_hPhysDevs.hBurst, result);

    DemoPrintTimeStamp(nTimeStamp);
    DemoLogf("Burst Recv:: Starts:%d, Completes:%d, Partial:=%d, Full=%d, Aborts:%d, Amount:%x, Errors:%d, OUErrors:%d, OUErrUser:%d, DMABusy:%d, Notify:%d, NotifyPartial:%d, FIFO Size:%d, Trigger:%d, Count:%d, Peak:%d, TriggerCount:%d\n",
        stat_Starts,
        stat_Completes,
        stat_Partial,
        stat_Full,
        stat_Aborts,
        stat_Amount,
        stat_Errors,
        stat_OUruns,
        stat_OUrunsUser,
        stat_DmaBusy,
        stat_Notify,
        stat_NotifyPartial,
        stat_FifoSize,
        stat_FifoTrigger,
        stat_FifoCount,
        stat_FifoPeak,
        stat_FifoTriggerCount
    );

    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_STAT_SEND_STARTS,             &stat_Starts            ); DemoCheckRc(g_hPhysDevs.hBurst, result);
    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_STAT_SEND_COMPLETES,          &stat_Completes         ); DemoCheckRc(g_hPhysDevs.hBurst, result);
    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_STAT_SEND_PARTIAL,            &stat_Partial           ); DemoCheckRc(g_hPhysDevs.hBurst, result);
    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_STAT_SEND_FULL,               &stat_Full              ); DemoCheckRc(g_hPhysDevs.hBurst, result);
    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_STAT_SEND_ABORTS,             &stat_Aborts            ); DemoCheckRc(g_hPhysDevs.hBurst, result);
    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_STAT_SEND_AMOUNT,             &stat_Amount            ); DemoCheckRc(g_hPhysDevs.hBurst, result);
    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_STAT_SEND_ERRORS,             &stat_Errors            ); DemoCheckRc(g_hPhysDevs.hBurst, result);
    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_STAT_SEND_OURUNS,             &stat_OUruns            ); DemoCheckRc(g_hPhysDevs.hBurst, result);
    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_STAT_SEND_OURUNS_USER,        &stat_OUrunsUser        ); DemoCheckRc(g_hPhysDevs.hBurst, result);
    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_STAT_SEND_DMABUSY,            &stat_DmaBusy           ); DemoCheckRc(g_hPhysDevs.hBurst, result);
    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_STAT_SEND_NOTIFY,             &stat_Notify            ); DemoCheckRc(g_hPhysDevs.hBurst, result);
    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_STAT_SEND_NOTIFY_PARTIAL,     &stat_NotifyPartial     ); DemoCheckRc(g_hPhysDevs.hBurst, result);
    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_SENDER_FIFO_SIZE,             &stat_FifoSize          ); DemoCheckRc(g_hPhysDevs.hBurst, result);
    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_SENDER_FIFO_TRIGGER,          &stat_FifoTrigger       ); DemoCheckRc(g_hPhysDevs.hBurst, result);
    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_SENDER_FIFO_COUNT,            &stat_FifoCount         ); DemoCheckRc(g_hPhysDevs.hBurst, result);
    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_SENDER_FIFO_PEAK,             &stat_FifoPeak          ); DemoCheckRc(g_hPhysDevs.hBurst, result);
    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_SENDER_FIFO_TRIGGER_COUNT,    &stat_FifoTriggerCount  ); DemoCheckRc(g_hPhysDevs.hBurst, result);

    DemoPrintTimeStamp(nTimeStamp);
    DemoLogf("Burst Send:: Starts:%d, Completes:%d, Partial:=%d, Full=%d, Aborts:%d, Amount:%x, Errors:%d, OUErrors:%d, OUErrUser:%d, DMABusy:%d, Notify:%d, NotifyPartial:%d, FIFO Size:%d, Trigger:%d, Count:%d, Peak:%d, TriggerCount:%d\n",
        stat_Starts,
        stat_Completes,
        stat_Partial,
        stat_Full,
        stat_Aborts,
        stat_Amount,
        stat_Errors,
        stat_OUruns,
        stat_OUrunsUser,
        stat_DmaBusy,
        stat_Notify,
        stat_NotifyPartial,
        stat_FifoSize,
        stat_FifoTrigger,
        stat_FifoCount,
        stat_FifoPeak,
        stat_FifoTriggerCount
    );

#ifdef STATS_SHOW_FRAME_REF
    result = OtxDrvGetAttributeValue(g_hPhysDevs.hBurst, OTX_ATTR_BURST_GET_FRAME_REF_DATA, &FrameRef, sizeof(FrameRef), &ui32); DemoCheckRc(g_hPhysDevs.hBurst, result);
    // just skip if we wrapped
    if (FrameRef.m_FrameCounter > nLastFrame) {
      DemoPrintTimeStamp(nTimeStamp);
      DemoLogf("Burst FRef:: %d fps  Cntr:%08x%08x  Time:%08x%08x ( ",
        nLastFrame == 0 ? 0 : (OTX_UINT32) (FrameRef.m_FrameCounter - nLastFrame),
        (OTX_UINT32) (FrameRef.m_FrameCounter >> 32),
        (OTX_UINT32) (FrameRef.m_FrameCounter),
        FrameRef.m_FrameDateTime.sUint32.nHigh,
        FrameRef.m_FrameDateTime.sUint32.nLow
      );
      DemoPrintTimeStamp(FrameRef.m_FrameDateTime);
      DemoLogf(")\n");
    }
    nLastFrame = FrameRef.m_FrameCounter;
#endif  // #ifdef STATS_SHOW_FRAME_REF

    OsaThreadSleep(1000);
  }

  DemoLogf("DemoStatsThread() terminating.\n");

  return 0;
}


/*******************************************************************************
  @func
    Burst Sender Function
*******************************************************************************/
void DemoBurstSender(OTX_UINT32 nSourceId)
{
  OTX_RESULT  otxRc;
  OTX_UINT32  ui32;


  // validate parameters
  if (nSourceId < 0 || nSourceId > 2) {
    DemoLogf("DemoBurstSender: Invalid nSourceId(%d)!\n", nSourceId);
    return;
  }

  if (!g_hLogicDevs.mBurstSender[nSourceId].bActive) {
    // we're not active, so just return
    return;
  }

START:
  if (g_hLogicDevs.mBurstSender[nSourceId].pTxFile) {
    // Prime the transmit buffer with 0xFF (only really needed for single TS transmission)
    memset(&g_hLogicDevs.mBurstSender[nSourceId].sqTxBuf[0], 0xFF, sizeof(g_hLogicDevs.mBurstSender[nSourceId].sqTxBuf) / sizeof(g_hLogicDevs.mBurstSender[nSourceId].sqTxBuf[0]));

    if (g_hLogicDevs.mBurstSender[nSourceId].nTsToSend == -1) {
      // All TS
      g_hLogicDevs.mBurstSender[nSourceId].cbTxBuf = fread(
        &g_hLogicDevs.mBurstSender[nSourceId].sqTxBuf,
        sizeof(g_hLogicDevs.mBurstSender[nSourceId].sqTxBuf[0]),
        sizeof(g_hLogicDevs.mBurstSender[nSourceId].sqTxBuf) / sizeof(g_hLogicDevs.mBurstSender[nSourceId].sqTxBuf[0]),
        g_hLogicDevs.mBurstSender[nSourceId].pTxFile
        );
      
    } else {
      // Single TS
      OTX_UINT8 singleTsBuf[DEMO_BURST_BUF_SIZE / 32]; // On some boards this is bigger than we actually use
      int cbSingleTsBuf;
      int byteCount = 0;
      int cbMaxSingleTsBuf = DEMO_BURST_BUF_SIZE / 32 / (g_demoEnv.coLi / g_demoEnv.coHighways) / sizeof(singleTsBuf[0]);
      int x;
      
      cbSingleTsBuf = fread(
        &singleTsBuf,
        sizeof(singleTsBuf[0]),
        cbMaxSingleTsBuf,
        g_hLogicDevs.mBurstSender[nSourceId].pTxFile
        );

      if (cbSingleTsBuf > 0) {
        if ((cbSingleTsBuf > 0) && (cbSingleTsBuf < cbMaxSingleTsBuf)) {
          // We are at the end of the file, and we don't have a full buffer to send.
          // Pad with 0xFF at the end of the buffer to make sure that we don't underrun the Burst Sender
          memset(&singleTsBuf[cbSingleTsBuf], 0xFF, (cbMaxSingleTsBuf - cbSingleTsBuf) / sizeof(singleTsBuf[0]) );
          cbSingleTsBuf = cbMaxSingleTsBuf;
        }
        
        // Dispurse the single TS to be transmitted in the whole transmit buffer 
        byteCount = 0;
        //DemoLogf("g_demoEnv.nCurrentLi=%d, g_demoEnv.coLi=%d, g_demoEnv.coHighways=%d, g_hLogicDevs.mBurstSender[nSourceId].nTsToSend=%d \n", g_demoEnv.nCurrentLi, g_demoEnv.coLi, g_demoEnv.coHighways, g_hLogicDevs.mBurstSender[nSourceId].nTsToSend);
        for ( x = (g_demoEnv.nCurrentLi % (g_demoEnv.coLi / g_demoEnv.coHighways)) + g_hLogicDevs.mBurstSender[nSourceId].nTsToSend; 
        (byteCount < cbSingleTsBuf) && (x < DEMO_BURST_BUF_SIZE); 
        x += (32 * (g_demoEnv.coLi / g_demoEnv.coHighways))
          ) {
          g_hLogicDevs.mBurstSender[nSourceId].sqTxBuf[x] = singleTsBuf[byteCount++];
        }
        
        assert(byteCount == cbSingleTsBuf);
          
      }

      g_hLogicDevs.mBurstSender[nSourceId].cbTxBuf = cbSingleTsBuf * 32 * (g_demoEnv.coLi / g_demoEnv.coHighways);

    }
    
    if (g_hLogicDevs.mBurstSender[nSourceId].cbTxBuf > 0) {
      ui32 = g_hLogicDevs.mBurstSender[nSourceId].cbTxBuf;

      assert(g_hLogicDevs.mBurstSender[nSourceId].cbTxBuf == DEMO_BURST_BUF_SIZE);

      otxRc = OtxBurstSendData(g_hPhysDevs.hBurst, nSourceId, g_hLogicDevs.mBurstSender[nSourceId].sqTxBuf, &ui32);
      DemoCheckRc(g_hPhysDevs.hBurst, otxRc);
      
      // check the amount buffered w/ how much we requested to be buffered
      //  (since we set the trigger amount to the buffer size, there should
      //    have been room to buffer this, so these two should be the same).
      //assert(ui32 == g_hLogicDevs.mBurstSender[nSourceId].cbTxBuf);
      if (ui32 != g_hLogicDevs.mBurstSender[nSourceId].cbTxBuf) {
        DemoLogf("DemoBurstSender[%d]: ui32 (%d) != g_hLogicDevs.mBurstSender[nSourceId].cbTxBuf (%d)\n", nSourceId, ui32 != g_hLogicDevs.mBurstSender[nSourceId].cbTxBuf);
      }
      
    } else {
      
      if (g_hLogicDevs.mBurstSender[nSourceId].bFileLoop) {
        if (g_hLogicDevs.mBurstSender[nSourceId].bVerbose) {
          DemoLogf("DemoBurstSender[%d]: Transmission complete. Starting file from the beginning.\n", nSourceId);
        }
        rewind(g_hLogicDevs.mBurstSender[nSourceId].pTxFile);
        goto START;
      } else {
        DemoLogf("DemoBurstSender[%d]: Transmission complete. Closing file\n", nSourceId);
        fclose(g_hLogicDevs.mBurstSender[nSourceId].pTxFile);
        g_hLogicDevs.mBurstSender[nSourceId].pTxFile = NULL;
        g_hLogicDevs.mBurstSender[nSourceId].bActive = OTX_FALSE;
      }
    }
    
  } else {
    // Send a text string (continously until aborted)

    // we MUST round amount to send to be bounded on the number of TS the highway hosts.
    //  (the buffer is representative of a burst DMA buffer for all TS on the highway hosted, TS0-TSn)
    g_hLogicDevs.mBurstSender[nSourceId].cbTxBuf = ((g_hLogicDevs.mBurstSender[nSourceId].cbTxBuf / (g_demoEnv.coLi / g_demoEnv.coHighways)) * (g_demoEnv.coLi / g_demoEnv.coHighways));
    assert(g_hLogicDevs.mBurstSender[nSourceId].cbTxBuf == DEMO_BURST_BUF_SIZE);

    ui32 = g_hLogicDevs.mBurstSender[nSourceId].cbTxBuf;
    otxRc = OtxBurstSendData(g_hPhysDevs.hBurst, nSourceId, g_hLogicDevs.mBurstSender[nSourceId].sqTxBuf, &ui32);
    DemoCheckRc(g_hPhysDevs.hBurst, otxRc);

    // check the amount buffered w/ how much we requested to be buffered
    //  (since we set the trigger amount to the buffer size, there should
    //    have been room to buffer this, so these two should be the same).
    assert(ui32 == g_hLogicDevs.mBurstSender[nSourceId].cbTxBuf);
  }

  return;
}

/*******************************************************************************
  @func
    Burst Receiver Function
*******************************************************************************/
void DemoBurstReceiver(OTX_UINT32 nSourceId)
{
  OTX_RESULT    otxRc;
  OTX_UINT8     buf[DEMO_BURST_BUF_SIZE];
  OTX_UINT32    ui32;
  OTX_UINT64    nFrameRefId;
  OTX_DATETIME  dtFrameRefTime;

  // validate parameters
  if (nSourceId < 0 || nSourceId > 2) {
    DemoLogf("DemoBurstReceiver: Invalid nSourceId(%d)...\n", nSourceId);
    return;
  }

  if (!g_hLogicDevs.mBurstReceiver[nSourceId].bActive) {
    // we're not active, so just return
    return;
  }

  // we want the frame reference id and time
  ui32 = sizeof(buf);
  otxRc = OtxBurstReadDataEx(g_hPhysDevs.hBurst, nSourceId, buf, &ui32, &nFrameRefId, &dtFrameRefTime);
  DemoCheckRc(g_hPhysDevs.hBurst, otxRc);

  // check the amount buffered w/ how much we requested to be buffered
  //  (since we set the trigger amount to the buffer size, there should
  //    have been room to buffer this, so these two should be the same).
  assert(ui32 == sizeof(buf));

  // echo if we're in verbose mode
  if (g_hLogicDevs.mBurstReceiver[nSourceId].bVerbose) {
    int x;

#ifdef DEMO_DISPLAY_BYTES // Display the first 14 bytes of from the current Li
    DemoPrintTimeStamp(dtFrameRefTime);

    // note: we only display the least significant 32 bits of the frame reference id.
    DemoLogf("BurstReceiver[%d][%d] (len=%.5d) ID:%08x ", nSourceId, g_demoEnv.nCurrentLi, sizeof(buf), (OTX_UINT32) (nFrameRefId & 0xffffffff));

    for (x = 0; (x < 14) && (x < (OTX_INT32) sizeof(buf)); x++) {
      // display the highway's currentLi's data
      DemoLogf("%.2X ", buf[x + ((g_demoEnv.nCurrentLi % (g_demoEnv.coLi / g_demoEnv.coHighways)) * 32)]);
    }
    DemoLogf("\n");
#else 
    if (g_demoEnv.eT1E1Mode == OTX_T1E1_MODE_T1) {
      // Display all of the F-bits of the current Li
      int bitCount = 0;
      DemoLogf("BurstReceiver[%d][%d] (len=%.4d) ", nSourceId, g_demoEnv.nCurrentLi, sizeof(buf));
      for (x = (g_demoEnv.nCurrentLi % (g_demoEnv.coLi / g_demoEnv.coHighways)); x < (OTX_INT32) sizeof(buf); x += 32) {
        // Display F-bit
        DemoLogf("%.1X", buf[x] & 0x01);
        bitCount++;
           
        // Group by 4 bits to make it easier to decypher the FAS bits (001011) within the multiframe
        if ((bitCount % 4) == 0) {
          DemoLogf(" ");
        }
      }
      DemoLogf("\n");
    } else {
      // Display the first bit of TS0 (sync timeslot) of the even frames, and then the odd frames
      // In this way we can more easily decypher the frame alignment pattern (001011xx).
      // It will be repeated in either the even or the odd frames.

      // Display even frames
      int byteCount = 0;
      DemoLogf("BurstReceiver[%d][%d] (len=%.4d) (1)", nSourceId, g_demoEnv.nCurrentLi, sizeof(buf));
      for (x = (g_demoEnv.nCurrentLi % (g_demoEnv.coLi / g_demoEnv.coHighways)); x < (OTX_INT32) sizeof(buf); x += 32) {
        // Display Sync bit of TS0 (MSB)
        if ((byteCount++ % 2) == 0) {
          continue;
        }
        DemoLogf("%.1X", (buf[x]>>7) & 0x01);

        // Group by 4 bytes to make it easier to decypher the synchronization bits (001011xx) within the multiframe
        if ((byteCount % 8) == 0) {
          DemoLogf(" ");
        }
      }
      DemoLogf("\n");
      
      // Display odd frames
      byteCount = 0;
      DemoLogf("BurstReceiver[%d][%d] (len=%.4d) (2)", nSourceId, g_demoEnv.nCurrentLi, sizeof(buf));
      for (x = (g_demoEnv.nCurrentLi % (g_demoEnv.coLi / g_demoEnv.coHighways)); x < (OTX_INT32) sizeof(buf); x += 32) {
        // Display Sync bit of TS0 (MSB)
        if ((byteCount++ % 2) == 1) {
          continue;
        }
        DemoLogf("%.1X", (buf[x]>>7) & 0x01);

        // Group by 4 bytes to make it easier to decypher the synchronization bits (001011xx) within the multiframe
        if (((byteCount+1) % 8) == 0) {
          DemoLogf(" ");
        }
      }
      DemoLogf("\n");

    }
#endif
  }

  // write data to a file if desired
  if (g_hLogicDevs.mBurstReceiver[nSourceId].pRxFile) {
    int cbWritten;

    if (g_hLogicDevs.mBurstReceiver[nSourceId].nTsToRecord == -1) {
      // All TS
      cbWritten = fwrite(buf, sizeof(buf[0]), sizeof(buf), g_hLogicDevs.mBurstReceiver[nSourceId].pRxFile);
      if (cbWritten != (int) sizeof(buf)) {
        DemoLogf("BurstReceiver[%d] Error: Could not write data to file.  (nFrameLen=%d, cbWritten=%d) \n", nSourceId, sizeof(buf), cbWritten);
      }
    } else {
      // Single TS
      OTX_UINT8 singleTsBuf[DEMO_BURST_BUF_SIZE/32];
      int byteCount = 0;
      int x;

      byteCount = 0;
      for (
          x = g_hLogicDevs.mBurstReceiver[nSourceId].nTsToRecord;
          x < (OTX_INT32) sizeof(buf); 
          x += (32 * (g_demoEnv.coLi / g_demoEnv.coHighways))
      ) {
        singleTsBuf[byteCount++] = buf[x];
      }
      cbWritten = fwrite(singleTsBuf, sizeof(singleTsBuf[0]), byteCount, g_hLogicDevs.mBurstReceiver[nSourceId].pRxFile);
      if (cbWritten != byteCount) {
        DemoLogf("BurstReceiver[%d] Error: Could not write data to file.  (byteCount=%d, cbWritten=%d) \n", nSourceId, byteCount, cbWritten);
      } else {
        //DemoLogf("BurstReceiver[%d] (byteCount=%d, cbWritten=%d) \n", nSourceId, byteCount, cbWritten);
      }
    }
  }
}

void DemoPrintEvent(
/*******************************************************************************
  @func
    Prints a Driver Event.
*******************************************************************************/
  OtxEventDataS *pEventData
)
{
  OTX_RESULT otxRc;
  char msgStr[256];

  switch (pEventData->m_eDeviceType) {
  case OTX_DEVICE_LI_T1E1:
    DemoPrintTimeStamp(pEventData->m_nTimeStamp);

    // Event from Li0 or Li1
    otxRc = OtxDrvEventCode2String(pEventData->m_hDevice, pEventData->m_nCode, 256, &msgStr[0]);
    if (OTX_SUCCESS(otxRc)) {
      if (pEventData->m_nParam) {
        DemoLogf("E1/T1[%d] Event: %s (%d)\n", pEventData->m_nSourceId, msgStr, pEventData->m_nParam);
      } else {
        DemoLogf("E1/T1[%d] Event: %s\n", pEventData->m_nSourceId, msgStr);
      }
    } else {
      DemoLogf("E1/T1[%d] Invalid Event: %ld \n", pEventData->m_nSourceId, pEventData->m_nCode);
    }
    break;

  case OTX_DEVICE_BURST:
    DemoPrintTimeStamp(pEventData->m_nTimeStamp);

    otxRc = OtxDrvEventCode2String(pEventData->m_hDevice, pEventData->m_nCode, 256, &msgStr[0]);
    if (OTX_SUCCESS(otxRc)) {
      DemoLogf("Burst[%d] Event: %s (%d)\n", pEventData->m_nSourceId, msgStr, pEventData->m_nParam);
    } else {
      DemoLogf("Burst[%d] Invalid Event: %ld \n", pEventData->m_nSourceId, pEventData->m_nCode);
    }
    break;

  default:
    // Unhandled event
    DemoPrintTimeStamp(pEventData->m_nTimeStamp);
    DemoLogf("Event from an unknown object 0x%X\n", pEventData->m_hDevice);
    break;
  }
}

OTX_INT32 DemoProcessUserCommand(char commandChar)
/*******************************************************************************

  @func

  Processes User Commands

  @comm

*******************************************************************************/
{
  OTX_RESULT result;

  DemoLogf("Key '%c' pressed\n", commandChar);

  switch (commandChar) {

  case '?':
  case 'h':
  case 'H':
    DemoLogf("Cmd  Synopsis\n");
    DemoLogf("===  ========\n");
    DemoLogf(" ?   Display Help text\n");
    DemoLogf(" h   Ditto\n");
    DemoLogf(" q   Quit\n");
    DemoLogf(" 0-7 Select active Li or Dsp, this Li/Dsp is then used for subsequent commands\n");
    DemoLogf(" a   Abort Burst[%d] Senders and Receivers\n",  g_demoEnv.nCurrentLi / (g_demoEnv.coLi / g_demoEnv.coHighways));
    DemoLogf(" r   Start Burst Receiver[%d] (no file logging)\n", g_demoEnv.nCurrentLi / (g_demoEnv.coLi / g_demoEnv.coHighways));
    DemoLogf(" R   Start Burst Receiver[%d] Prompt for file to recv\n", g_demoEnv.nCurrentLi / (g_demoEnv.coLi / g_demoEnv.coHighways));
    DemoLogf(" s   Start Burst Sender[%d]. Continously send a text string\n", g_demoEnv.nCurrentLi / (g_demoEnv.coLi / g_demoEnv.coHighways));
    DemoLogf(" S   Start Burst Sender[%d]. Prompt for file to send\n", g_demoEnv.nCurrentLi / (g_demoEnv.coLi / g_demoEnv.coHighways));
    DemoLogf(" L   Burst Sender[%d]. Toggle loop file sends (currently %s)\n", g_demoEnv.nCurrentLi / (g_demoEnv.coLi / g_demoEnv.coHighways), g_hLogicDevs.mBurstSender[g_demoEnv.nCurrentLi / (g_demoEnv.coLi / g_demoEnv.coHighways)].bFileLoop ? "On" : "Off");
    DemoLogf(" e   Enable Burst\n");
    DemoLogf(" d   Disable Burst\n");
    DemoLogf(" v   Toggle verbose (print received data) setting Burst Receiver[%d]\n", g_demoEnv.nCurrentLi / (g_demoEnv.coLi / g_demoEnv.coHighways));
    DemoLogf(" V   Toggle verbose (print send progress) setting Burst Sender[%d]\n", g_demoEnv.nCurrentLi / (g_demoEnv.coLi / g_demoEnv.coHighways));
    DemoLogf(" l   Query Line Status of both Lis\n");
    DemoLogf(" t   Toggle verbose buffer usage statistics\n");
    DemoLogf(" f   Forcing resynchronization\n");
    DemoLogf(" F   Reset all lines\n");
    DemoLogf(" T   Reset buffer usage statistics\n");
    DemoLogf(" D   Dump buffer usage statistics\n");
    DemoLogf(" n   Set burst notification value (1-32)\n");
    break;

  case 'e':
    // by enabling the Burst device, we start the process of transferring data
    // from the hardware to the driver via DMA.  The data is stored in the Burst
    // device's buffer for later processing to connected logical devices.
    DemoLogf("Enable g_hPhysDevs.hBurst\n");
    result = OtxDrvEnable(g_hPhysDevs.hBurst, OTX_TASK_SYNC, 5000L);
    DemoCheckRc(g_hPhysDevs.hBurst, result);
    break;

  case 'd':
    // by disabling the Burst device, we stop the process of transferring data
    // from the hardware to the driver via DMA.
    DemoLogf("Disable g_hPhysDevs.hBurst\n");
    result = OtxDrvDisable(g_hPhysDevs.hBurst, OTX_TASK_SYNC, 5000L);
    DemoCheckRc(g_hPhysDevs.hBurst, result);
    break;

  case 't':
    g_bStatsThreadActive = !g_bStatsThreadActive;
    if (g_bStatsThreadActive) {
      DemoLogf("Starting Stats Thread\n");
      OsaThreadStart(DemoStatsThread);
    }
    break;

  case 'f':
    {
      OTX_INT32 nLi;

      for (nLi = 0; nLi < g_demoEnv.coLi; nLi++) {
        DemoLogf("Forcing resynchronization on Li #%d... ", nLi);
        result = OtxT1E1ForceResynch(g_hPhysDevs.hLi[nLi]); DemoCheckRc(g_hPhysDevs.hLi[nLi], result);
        DemoLogf("Done.");
      }
    }
    break;

  case 'F':
    {
      OTX_INT32 nLi;

      for (nLi = 0; nLi < g_demoEnv.coLi; nLi++) {
        DemoLogf("Reseting Li #%d... ", nLi);
        result = OtxDrvReset(g_hPhysDevs.hLi[nLi]); DemoCheckRc(g_hPhysDevs.hLi[nLi], result);
        DemoLogf("Done.");
      }
    }
    break;

  case 'v':
    g_hLogicDevs.mBurstReceiver[g_demoEnv.nCurrentLi / (g_demoEnv.coLi / g_demoEnv.coHighways)].bVerbose = !g_hLogicDevs.mBurstReceiver[g_demoEnv.nCurrentLi / (g_demoEnv.coLi / g_demoEnv.coHighways)].bVerbose;
    DemoLogf("Setting verbose setting to %s for Burst Receiver[%d] Thread\n", (g_hLogicDevs.mBurstReceiver[g_demoEnv.nCurrentLi / (g_demoEnv.coLi / g_demoEnv.coHighways)].bVerbose)?"ON":"OFF", g_demoEnv.nCurrentLi / (g_demoEnv.coLi / g_demoEnv.coHighways));
    break;

  case 'V':
    g_hLogicDevs.mBurstSender[g_demoEnv.nCurrentLi / (g_demoEnv.coLi / g_demoEnv.coHighways)].bVerbose = !g_hLogicDevs.mBurstSender[g_demoEnv.nCurrentLi / (g_demoEnv.coLi / g_demoEnv.coHighways)].bVerbose;
    DemoLogf("Setting verbose setting to %s for Burst Sender[%d] Thread\n", (g_hLogicDevs.mBurstSender[g_demoEnv.nCurrentLi / (g_demoEnv.coLi / g_demoEnv.coHighways)].bVerbose)?"ON":"OFF", g_demoEnv.nCurrentLi / (g_demoEnv.coLi / g_demoEnv.coHighways));
    break;

  case 'a':
    {
      OTX_UINT32 nHighwayNo;

      nHighwayNo = g_demoEnv.nCurrentLi / (g_demoEnv.coLi / g_demoEnv.coHighways);

      if (g_hLogicDevs.mBurstReceiver[nHighwayNo].pRxFile) {
        DemoLogf("DemoBurstReceiver[%d]: Closing file\n", nHighwayNo);
        fclose(g_hLogicDevs.mBurstReceiver[nHighwayNo].pRxFile);
        g_hLogicDevs.mBurstReceiver[nHighwayNo].pRxFile = NULL;
      }
      g_hLogicDevs.mBurstReceiver[nHighwayNo].bActive = OTX_FALSE;

      if (g_hLogicDevs.mBurstReceiver[nHighwayNo].pRxFile) {
        DemoLogf("mBurstSender[%d]: Closing file\n", nHighwayNo);
        fclose(g_hLogicDevs.mBurstSender[nHighwayNo].pTxFile);
        g_hLogicDevs.mBurstSender[nHighwayNo].pTxFile = NULL;
      }
      g_hLogicDevs.mBurstSender[nHighwayNo].bActive   = OTX_FALSE;
    }
    break;

  case 'r':
    {
      OTX_UINT32 nHighwayNo;

      nHighwayNo = g_demoEnv.nCurrentLi / (g_demoEnv.coLi / g_demoEnv.coHighways);
      if (g_hLogicDevs.mBurstReceiver[nHighwayNo].bActive == OTX_FALSE) {
        g_hLogicDevs.mBurstReceiver[nHighwayNo].bActive = OTX_TRUE;
      } else if (g_hLogicDevs.mBurstReceiver[nHighwayNo].pRxFile) {
        DemoLogf("DemoBurstReceiver[%d]: Closing file\n", nHighwayNo);
        fclose(g_hLogicDevs.mBurstReceiver[nHighwayNo].pRxFile);
        g_hLogicDevs.mBurstReceiver[nHighwayNo].pRxFile = NULL;
      } else {
        DemoLogf("Burst Receiver[%d] already running.\n", nHighwayNo);
      }
    }
    break;

  case 'R':
    {
      char sqzBuf[OTX_MAX_PATH];
      OTX_UINT32 nHighwayNo;
      static OTX_INT32 nTsToRecord = -1;

      nHighwayNo = g_demoEnv.nCurrentLi / (g_demoEnv.coLi / g_demoEnv.coHighways);
      if (g_hLogicDevs.mBurstReceiver[nHighwayNo].bActive == OTX_FALSE || g_hLogicDevs.mBurstReceiver[nHighwayNo].pRxFile == NULL) {
        DemoLogf("BurstReceiver[%d]. Please enter file to receive: ", nHighwayNo);
        OsaConsoleGetStr(sqzBuf);
        if (sqzBuf[0] == '\0') {
          DemoLogf("File name not entered. Operation aborted\n");
          break;
        }

        if ((g_hLogicDevs.mBurstReceiver[nHighwayNo].pRxFile = fopen(sqzBuf, "wb")) == NULL) {
          DemoLogf("Error: Could not open file %s \n", sqzBuf);
          break;
        }

        DemoLogf("BurstReceiver[%d]. Please enter single TS[0-%d] to receive to file (-1=All) [%d]: ", nHighwayNo, (32 * (g_demoEnv.coLi / g_demoEnv.coHighways))-1, nTsToRecord);
        OsaConsoleGetStr(sqzBuf);
        if (sqzBuf[0]) {
          nTsToRecord = strtol(sqzBuf, NULL, 0);
        }
        g_hLogicDevs.mBurstReceiver[nHighwayNo].nTsToRecord = nTsToRecord;

        g_hLogicDevs.mBurstReceiver[nHighwayNo].bActive = OTX_TRUE;
      } else {
        DemoLogf("Burst Receiver[%d] Thread already running.\n", nHighwayNo);
      }
    }
    break;

  case 's':
    {
      OTX_UINT32 nHighwayNo, i, base;

      nHighwayNo = g_demoEnv.nCurrentLi / (g_demoEnv.coLi / g_demoEnv.coHighways);
      if (g_hLogicDevs.mBurstSender[nHighwayNo].bActive == OTX_FALSE) {
        DemoLogf("BurstSender[%d]. Continously send static values\n", nHighwayNo);

        g_hLogicDevs.mBurstSender[nHighwayNo].pTxFile = NULL;

        // Fill sqTxBuf with the text string (as many copies of the text string that will fit in sqTxBuf).
        g_hLogicDevs.mBurstSender[nHighwayNo].cbTxBuf = 0;
        for (i = 0; i < DEMO_BURST_BUF_SIZE; i++) {
          // buffer is laid out as (allows easy verification of data):
          //   LI0 [0]=0x00 [1]=0x01 .. [10]=0x00 [11]=0x01 .. [30]=0x00 [31]=0x01
          //   LI1 [0]=0x10 [1]=0x11 .. [10]=0x10 [11]=0x11 .. [30]=0x10 [31]=0x11
          //      ...
          //   LI6 [0]=0x60 [1]=0x61 .. [10]=0x60 [11]=0x61 .. [30]=0x60 [31]=0x61
          //   LI7 [0]=0x70 [1]=0x71 .. [10]=0x70 [11]=0x71 .. [30]=0x70 [31]=0x71
    
          base = (((g_demoEnv.coLi / g_demoEnv.coHighways) * nHighwayNo) * 0x10);
          g_hLogicDevs.mBurstSender[nHighwayNo].sqTxBuf[i] = (char) (base + (((i % ((g_demoEnv.coLi / g_demoEnv.coHighways) * 32)) / 32) * 0x10) + ((i % 32) % 10));
          g_hLogicDevs.mBurstSender[nHighwayNo].bActive = OTX_TRUE;
        }
        g_hLogicDevs.mBurstSender[nHighwayNo].cbTxBuf = DEMO_BURST_BUF_SIZE;
      } else {
        DemoLogf("Burst Sender[%d] already running.\n", nHighwayNo);
      }
    }

    break;

  case 'S':
    {
      char sqzBuf[OTX_MAX_PATH];
      OTX_UINT32 nHighwayNo;
      static OTX_INT32 nTsToSend = -1;

      nHighwayNo = g_demoEnv.nCurrentLi / (g_demoEnv.coLi / g_demoEnv.coHighways);
      if (g_hLogicDevs.mBurstSender[nHighwayNo].bActive == OTX_FALSE) {
        DemoLogf("BurstSender[%d]. Please enter file to send: ", nHighwayNo);

        OsaConsoleGetStr(sqzBuf);
        if (sqzBuf[0] == '\0') {
          DemoLogf("File name not entered. Operation aborted\n");
          break;
        }

        if ((g_hLogicDevs.mBurstSender[nHighwayNo].pTxFile = fopen(sqzBuf, "rb")) == NULL) {
          DemoLogf("Error: Could not open file %s \n", sqzBuf);
          break;
        }

        DemoLogf("BurstSender[%d]. Please enter single TS[0-%d] to send from file (-1=All) [%d]: ", nHighwayNo, (32 * (g_demoEnv.coLi / g_demoEnv.coHighways))-1, nTsToSend);
        OsaConsoleGetStr(sqzBuf);
        if (sqzBuf[0]) {
          nTsToSend = strtol(sqzBuf, NULL, 0);
        }
        g_hLogicDevs.mBurstSender[nHighwayNo].nTsToSend = nTsToSend;

        g_hLogicDevs.mBurstSender[nHighwayNo].bActive = OTX_TRUE;
        g_hLogicDevs.mBurstSender[nHighwayNo].bVerbose = OTX_TRUE;
      } else {
        DemoLogf("Burst Sender[%d] already running.\n", nHighwayNo);
      }
    }
    break;

  case 'L':
    g_hLogicDevs.mBurstSender[g_demoEnv.nCurrentLi / (g_demoEnv.coLi / g_demoEnv.coHighways)].bFileLoop = !g_hLogicDevs.mBurstSender[g_demoEnv.nCurrentLi / (g_demoEnv.coLi / g_demoEnv.coHighways)].bFileLoop;
    DemoLogf("Setting loop file sends %s for Burst Sender[%d] Thread\n", (g_hLogicDevs.mBurstSender[g_demoEnv.nCurrentLi / (g_demoEnv.coLi / g_demoEnv.coHighways)].bFileLoop)?"ON":"OFF", g_demoEnv.nCurrentLi / (g_demoEnv.coLi / g_demoEnv.coHighways));
    break;

  case 'l':
    {
      OtxT1E1Layer1StatusE eLayer1Status;
      OTX_INT32 nLi;

      for (nLi = 0; nLi < g_demoEnv.coLi; nLi++) {
        DemoLogf("Line status of Li #%d is ", nLi);
        result = OtxT1E1GetLayer1Status(g_hPhysDevs.hLi[nLi], &eLayer1Status); DemoCheckRc(g_hPhysDevs.hLi[nLi], result);
        DemoLogf("%s\n", eLayer1Status == OTX_T1E1_LAYER1_OK ? "OK" : "DOWN");
      }
    }
    break;

  case '0':
  case '1':
  case '2':
  case '3':
  case '4':
  case '5':
  case '6':
  case '7':
    if (commandChar - '0' < g_demoEnv.coLi) {
      // Select current Li
      g_demoEnv.nCurrentLi = commandChar - '0';
      DemoLogf("LI#%d Selected ....\n", g_demoEnv.nCurrentLi);
    }
    break;

  case 'T':
    // reset stats
    OtxDrvDbgDump(g_hPhysDevs.hBurst, 101);
    break;

  case 'D':
    // dump stats to debugger
    OtxDrvDbgDump(g_hPhysDevs.hBurst, 100);
    break;

  case 'n':
    {
      OTX_UINT32 nValue;
      char sValue[80];

      result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_TRIGGER, &nValue); DemoCheckRc(g_hPhysDevs.hBurst, result);
      DemoLogf("Enter queue notification value (1-32) [currently %d]: ", nValue);
      OsaConsoleGetStr(sValue);
      if (sValue[0] == '\0') {
        break;
      }
      nValue = strtol(sValue, NULL, 0);

      // before setting the new trigger value, the burst device must not be enabled
      result = OtxDrvDisable(g_hPhysDevs.hBurst, OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hPhysDevs.hBurst, result);
      result = OtxDrvSetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_TRIGGER, nValue); DemoCheckRc(g_hPhysDevs.hBurst, result);
      result = OtxDrvEnable(g_hPhysDevs.hBurst,  OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hPhysDevs.hBurst, result);

      result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_TRIGGER, &nValue); DemoCheckRc(g_hPhysDevs.hBurst, result);
      DemoLogf("New queue notification value is %d\n", nValue);

      break;
    }

  case 'q':
  case 'Q':
    return 1;
    break;

  case '~':
    {
      char strBuf[12];
      OTX_INT32 nSleepTime;

      DemoPrintf("Enter milliseconds to sleep :");
      OsaConsoleGetStr(strBuf);

      if (strBuf[0]) {
        nSleepTime = strtol(strBuf, NULL, 0);
        DemoPrintf("Sleeping %d milliseconds\n", nSleepTime);
        OsaThreadSleep(nSleepTime);
      }
    }
    break;

  default:
    DemoLogf("Unknown command. Press '?' for help, 'q' to quit\n");
    break;
  }

  return 0;
}




/*******************************************************************************

  @func

  Parse the program's command line arguments and calls upon
  requested actions.

  There are some optional switches that can be used following
  the driver name on the command line. Switches are are
  separated by spaces, and can appear in arbitrary order.
  Switches consist of a hyphen and a letter:

  @comm

*******************************************************************************/
static void DemoParseCmdLineArgs(int argc, char *argv[], DemoEnvT *envP)
{
  short  i;

  // Set reasonable defaults before reading from command line
  envP->pszInputStr       = NULL;
  envP->bExitOnError      = OTX_FALSE;
  envP->eT1E1Mode         = OTX_T1E1_MODE_E1;
  envP->boardNo           = DEMO_DEFAULT_BOARD_NO;
  envP->nBoardType        = OTX_DEVICE_THOR_8_PCI_PLUS;
  envP->coLi              = 8;
  envP->coHighways        = 2;
  envP->nCurrentLi        = 0;
  envP->bDisableBurstTx   = OTX_FALSE;
  envP->bDisableBurstRx   = OTX_FALSE;

  strcpy(envP->logFileName, envP->cwd);
  strcat(envP->logFileName, DEMO_DEFAULT_LOG_FILE);
  for (i = 1; i < argc; i++) {        // Scan command line arguments
    if (*(argv[i]+1) == 'Z') {   // Command input string -- case sensitive
      envP->pszInputStr = argv[i]+2;
    } else {
      OsaStdStrUpr(argv[i]);
      if (*argv[i] == '-' || *argv[i] == '/') {  // Is argument a legal switch?
        if (*(argv[i]+1) == 'L') {   // Log file
          strcpy(envP->logFileName, argv[i] + 2);
        } else if (*(argv[i]+1) == 'Q') {   // Exit on error
          envP->bExitOnError = OTX_TRUE;
        } else if (*(argv[i]+1) == 'B') {   // Board number
          envP->boardNo = strtoul(argv[i]+2, NULL, 0);
        } else if (*(argv[i]+1) == 'T') {   // Board Type
          if ( *(argv[i]+2) == '4') {
            envP->nBoardType = OTX_DEVICE_THOR_4_PCI_PLUS;
            envP->coLi = 4;
            envP->coHighways = 1;
          } else if ( *(argv[i]+2) == 'P') {
            envP->nBoardType = OTX_DEVICE_THOR_8_PCI_PLUS;
            envP->coLi = 8;
            envP->coHighways = 2;
          } else if ( *(argv[i]+2) == 'T') {
            envP->nBoardType = OTX_DEVICE_THOR_2_PCI_PLUS;
            envP->coLi = 2;
            envP->coHighways = 2;
          } else if ( *(argv[i]+2) == 'S') {
            envP->nBoardType = OTX_DEVICE_SLEIPNIR_1_PCI;
            envP->coLi = 1;
            envP->coHighways = 1;
          } else if ( *(argv[i]+2) == 'H') {
            envP->nBoardType = OTX_DEVICE_HUGIN_2_PC104;
            envP->coLi = 2;
            envP->coHighways = 2;
          } else if ( *(argv[i]+2) == '1') {
            envP->nBoardType = OTX_DEVICE_THOR_1_PCI_PLUS;
            envP->coLi = 1;
            envP->coHighways = 1;
          } else {
            DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
            DemoHelpPanel();
            OsaConsoleExit();
            exit(1);
          }

      } else if (*(argv[i]+1) == 'M') {   // Li Mode
          if (*(argv[i]+2) == 'E') {   // E1
            envP->eT1E1Mode    = OTX_T1E1_MODE_E1;
          } else if (*(argv[i]+2) == 'T') {   // T1
            envP->eT1E1Mode    = OTX_T1E1_MODE_T1;
          } else {
            DemoLogf("Error: Invalid Li Mode (E1/T1) command line switch %s\n", argv[i]);
            DemoHelpPanel();
            OsaConsoleExit();
            exit(1);
          }
      } else if (*(argv[i]+1) == 'D') {   // Disable one direction of burst
          if (*(argv[i]+2) == 'T') {   // TX
            envP->bDisableBurstTx = OTX_TRUE;
          } else if (*(argv[i]+2) == 'R') {   // T1
            envP->bDisableBurstRx = OTX_TRUE;
          } else {
            DemoLogf("Error: Invalid Burst direction-disabling command line switch %s\n", argv[i]);
            DemoHelpPanel();
            OsaConsoleExit();
            exit(1);
          }
        } else {
          DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
          DemoHelpPanel();
          OsaConsoleExit();
          exit(1);
        }
      } else {
        DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
        DemoHelpPanel();
        OsaConsoleExit();
        exit(1);
      }
    }
  }
}


/*******************************************************************************

  @func

  Displays a help for the command line.

  @comm

*******************************************************************************/
static void DemoHelpPanel()
{
  DemoLogf("\nCommand line switches:\n");
  DemoLogf(" /L<fileName> : Log file name\n");
  DemoLogf(" /B<boardNo>  : Board number (Default is %d)\n", DEMO_DEFAULT_BOARD_NO);
  DemoLogf(" /T[T|4|P|S]    : Board Type used for the demo, where\n");
  DemoLogf("                  T = Use Thor-2-PCI-Plus Board\n");
  DemoLogf("                  4 = Use Thor-4-PCI-Plus Board\n");
  DemoLogf("                  P = Use Thor-8-PCI-Plus Board\n");
  DemoLogf("                  S = Use Sleipnir Board\n");
  DemoLogf("                  H = Use Hugin-2-PC104 Board\n");
  DemoLogf("                  1 = Use Thor-1-PCI-Plus Board\n");
  DemoLogf(" /ME          : E1 Li mode\n");
  DemoLogf(" /MT          : T1 Li mode\n");
  DemoLogf(" /D[T|R]      : Disable one direction of the Burst device\n");
  DemoLogf("                  T = Disable Burst TX\n");
  DemoLogf("                  R = Disable Burst RX\n");
  DemoLogf(" /Q           : Exit without prompting on errors\n");
}


/*******************************************************************************

  @func

  Displays a message on the screen and prints it to the logfile.

  @comm

*******************************************************************************/
static void DemoLogf(char *formatStr, ...)
{
  va_list argList;

  //EnterCriticalSection(&g_demoLogfCritSec);

  va_start(argList, formatStr);

  OsaConsolePrintF(formatStr, argList);
  if (g_pDemoLogFile) vfprintf(g_pDemoLogFile,   formatStr, argList);

  va_end(argList);

  //LeaveCriticalSection(&g_demoLogfCritSec);
}


/*******************************************************************************
*******************************************************************************/
void DemoAtExit(void)
{
  OTX_INT32 i;

  for (i = 0; i < g_demoEnv.coHighways; i++) {
    if (g_hLogicDevs.mBurstSender[i].pTxFile) {
      fclose(g_hLogicDevs.mBurstSender[i].pTxFile);
      g_hLogicDevs.mBurstSender[i].pTxFile = NULL;
    }
    if (g_hLogicDevs.mBurstReceiver[i].pRxFile) {
      fclose(g_hLogicDevs.mBurstReceiver[i].pRxFile);
      g_hLogicDevs.mBurstReceiver[i].pRxFile = NULL;
    }
  }

  if (g_pDemoLogFile) {
    fclose(g_pDemoLogFile);
  }
}

int main(int argc, char *argv[])
/*******************************************************************************
*******************************************************************************/
{
  OTX_INT32 exitFlag = 0;
  OTX_RESULT result;  // OTX Standard Return Code
  char commandChar;

  OsaConsoleInit();

  printf("T8pBurstEx, %s, Odin TeleSystems Inc.\n", __DATE__);
  printf("Example program for Burst device.\n");

  DemoParseCmdLineArgs(argc, argv, &g_demoEnv);

  if (g_demoEnv.pszInputStr) {
    OsaConsoleInput(g_demoEnv.pszInputStr);
  }

  if ((g_pDemoLogFile = fopen(g_demoEnv.logFileName, "wt")) == NULL) {
    printf("Couldn't open logfile %s\n", g_demoEnv.logFileName);
//    OsaConsoleExit();
//    exit(-1);
  }
  atexit(DemoAtExit);

  DemoLogf("Log file is %s\n", g_demoEnv.logFileName);

  // Query Library revision
  {
    char szVersionInfo[255];
    OTX_BOOL bIsStaticLinkedLibrary;
    char szLibFileName[64];
    result = OtxDrvGetAttributeValueSz(OTX_LIB_HANDLE, OTX_ATTR_LIB_REV_STR, szVersionInfo, 255); DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);
    result = OtxDrvGetAttributeValueBool(OTX_LIB_HANDLE, OTX_ATTR_LIB_TYPE_STATIC, &bIsStaticLinkedLibrary); DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);
    result = OtxDrvGetAttributeValueSz(OTX_LIB_HANDLE, OTX_ATTR_LIB_FILE_NAME, szLibFileName, 64); DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);

    DemoLogf("Library version %s (filename %s, %s linked)\n", szVersionInfo, szLibFileName, bIsStaticLinkedLibrary ? "static":"dynamic");
  }


  // Connect to the Driver
  DemoLogf("Connecting lib...\n");
  result = OtxDrvConnectLib();
  switch (OTX_RESULT_CODE(result)) {
  case OTX_RES_HANDLE:
    DemoLogf("Error: Could not connect to the OTX library.\nPossible cause: OTX Driver not loaded\n");
    OsaConsoleExit();
    OtxDrvReleaseLib();
    exit(1);
    break;
  case OTX_RES_NOT_SUPPORTED:
    DemoLogf("Error: Mismatch between OTX Driver revision and OTX Library revision.\nRevisions must match.\n");
    OsaConsoleExit();
    OtxDrvReleaseLib();
    exit(1);
    break;
  default:
    DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);
    break;
  }

  // Create Event Queues for Driver --> Application Communications
  DemoLogf("DemoCreateEventQueue\n");
  DemoCreateEventQueue();

  // Connect (get handles) to the physical devices on the Thor-2-PCMCIA board
  DemoLogf("DemoCreatePhysicalDevices\n");
  DemoCreatePhysicalDevices();

  // Configure E1/T1 devices
  DemoLogf("Configuring E1/T1 devices...\n");
  DemoConfigureE1T1Devices();
  DemoLogf("Configured E1/T1 devices\n");

  // Enable physical devices. Their device state should now be OTX_DEVICE_WORKING
  DemoLogf("DemoEnablePhysicalDevices\n");
  DemoEnablePhysicalDevices();

  // Set clock source for the board
  result = OtxBrdSetClocks(g_hPhysDevs.hThorController, OTX_CLOCK_MODE_LOCAL, OTX_CLOCK_SOURCE_INTERNAL); DemoCheckRc(g_hPhysDevs.hThorController, result);

  // Create and get handles to the Pipe devices needed by this application
  DemoLogf("DemoConnectDevices\n");
  DemoConnectDevices();

  // Main Loop. Loop here and wait for Events from the driver or commands from the user ....
  OsaThreadStart(DemoIdleFunction);

  OsaThreadSleep(300); // Only so the following message will show on the screen

  DemoLogf("\nPress 'h' for help\n");

  while (!exitFlag) {
    commandChar = OsaConsoleGetCh();
    exitFlag = DemoProcessUserCommand(commandChar);
  }

  // abort threads
  DemoLogf("\nExiting...\n");
  {
    OTX_INT32 i;


    g_bStatsThreadActive = OTX_FALSE;
    for (i = 0; i < g_demoEnv.coHighways; i++) {
      g_hLogicDevs.mBurstReceiver[i].bActive = OTX_FALSE;
      g_hLogicDevs.mBurstSender[i].bActive = OTX_FALSE;
    }

    OsaThreadSleep(1500);   // give threads a chance to exit
  }

  // Clean up
  OsaConsoleExit();
  OtxDrvReleaseLib();

  return 0;
}
